/* dpaccess_pass
   written by Paolo Subiaco iw3grx@ir3ip.ampr.org
   Control, en/decode passwords to authenticate dpaccess users
   (small Tcl script to access dpbox features through Common Gateway Interface)
   Also, insert new user information in the password_file (but you must call dpaccess_pass as root)
   Compile with gcc -O2 -lcrypt -o dpaccess_pass dpaccess_pass.c

   Please adjust the two parameters below:
   password_file : full path of the file within user informations (callsign, password, name, etc)
   httpd_uid : User Identify of the http daemon (normally "nobody" , which has often tUID=99)
*/

#define password_file "/var/spool/dpbox/.cgi/info"
// #define httpd_uid     48
#define httpd_uid     99

#include <crypt.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#define _XOPEN_SOURCE
#define __USE_XOPEN
#include <unistd.h>

int main(int argc, char **argv) {
  char encr[32];
  char *pass;
  char salt[3];
  int fd;
  FILE *st;
  char buffer[128];
  int i=0,j=0,found=0;
  int uid=1000;

  uid=getuid();
  if (uid==httpd_uid) {
    // dpaccess_pass is called by http server
    if (argc!=4) { 
      printf("Error: invalid number of parameters");
      return -1;
    }
    if (!strcmp(argv[1],"control")) {	
      /* sintax is dpaccess_pass control CALLSIGN PLAIN_TEXT_PASSWORD
       * Return in the stdout
       *	 Ok		-> The password is right!
       *	No-User 	-> The file does not contain that callsign
       *	No-Pass		-> Provided password don't match with the specified
       */
      if ((st=fopen(password_file,"r"))==NULL) {	
	printf("Error: cannot open password file");
	return -2;
      }
      while (fgets(buffer,128,st)) { 
	found=0;
	for (i=0;i<=6;i++) {
	  if (buffer[i]==':' && argv[2][i]==0) {
	    found=1;
	    break;
	  }
	  if (buffer[i]!=argv[2][i])
	    break;
	}
	if (found)
	  break;
      }
      fclose(st);
      if (!found) {
	printf("No-User");
	return 0;
      }
      else { 
	// i points to the field separator ':'
	for (j=0,i++;buffer[i]!=':';j++,i++)
	  encr[j]=buffer[i];
	encr[j]=0;
	strncpy(salt,encr,2);
        pass=crypt(argv[3],salt);
	if (!strcmp(encr,pass)) {
	  printf("Ok ");
	  for (i++;buffer[i]!=':';i++)
	    putchar(buffer[i]);
	  }
	else
	  printf("No-Pass");
	
	return 0;
      }
    }
    else if (!strcmp(argv[1],"encode")) {
      /* dpaccess_pass encode callsign
       *   Return in the stdout the encrypt password of callsign: this password will be stored in the cookie
       *   of callsign browser, and must be decode from the next dpaccess_pass (using the next instruction) to
       *   grant access
       */
      // Now it is simply equivalent to the DES encrypt password without salt: it must be changed in the future!
      if ((st=fopen(password_file,"r"))==NULL) {	
	printf("Error: cannot open password file");
	return -2;
      }
      while (fgets(buffer,128,st)) { 
	found=0;
	for (i=0;i<=6;i++) {
	  if (buffer[i]==':' && argv[2][i]==0) {
	    found=1;
	    break;
	  }
	  if (buffer[i]!=argv[2][i])
	    break;
	}
	if (found)
	  break;
      }
      fclose(st);
      if (!found) {
	printf("No-User");
	return 0;
      }
      else { 
	// i points to the field separator ':'
	for (j=0,i++;buffer[i]!=':';j++,i++)
	  encr[j]=buffer[i];
	encr[j]=0;
	printf("%s",encr+2);
	return 0;
      }
    }
    else if (!strcmp(argv[1],"decode")) {
      /* dpaccess_pass decode callsign encr_password_without_salt
         Compare the cookie password and the one stored in the DPBOX_ACCESS_FILE.
         Return
	 "Ok NAME"	        -> Access grant
	 No-User		-> The DPBOX_ACCESS_FILE does not contain that callsign
	 No-Pass		-> The password provided is wrong
      */
      if ((st=fopen(password_file,"r"))==NULL) {	
	printf("Error: cannot open password file");
	return -2;
      }
      while (fgets(buffer,128,st)) { 
	found=0;
	for (i=0;i<=6;i++) {
	  if (buffer[i]==':' && argv[2][i]==0) {
	    found=1;
	    break;
	  }
	  if (buffer[i]!=argv[2][i])
	    break;
	}
	if (found)
	  break;
      }
      fclose(st);
      if (!found) {
	printf("No-User");
	return 0;
      }
      else { 
	// i points to the field separator ':'
	for (j=0,i++;buffer[i]!=':';j++,i++)
	  encr[j]=buffer[i];
	encr[j]=0;
	if (!strcmp(encr+2,argv[3])) {
	  printf("Ok ");
	  for (i++;buffer[i]!=':';i++) // extract user NAME
	    putchar(buffer[i]);
	  }
	else
	  printf("No-Pass");
	return 0;
      }
    }
    return -4;
  }
  else if (uid==0) {
    // dpaccess_pass is called by root : root can add new users
    printf("Running program as root..\n");
    if (argc!=5 || strcmp(argv[1],"add")) {
      printf ("ERROR: you must call program as\n%s add callsign password Name\nFor example %s add iw3grx i5f3adF Paolo",\
	      argv[0],argv[0]);
      return -5;
    }
    // open /dev/random and read 2 byte to use as the salt
    if ((fd=open("/dev/random",O_RDONLY))<0) {
      printf("Error opening /dev/random file to calculate salt...using another alghoritm\n");
      salt[0]=(char)(~argv[2][2]*argv[2][3]*argv[3][0]);
      salt[1]=(char)(~argv[3][1]*argv[3][3]*~argv[4][0]);
      salt[2]=0;
    } else {
      read(fd,salt,2);
      salt[2]=0;
      close(fd);
    }
    printf("Opening password file...\n");
    if ((st=fopen(password_file,"r+"))==NULL) {
      // password file must be created!
      printf("Password file does not exist: create it!\n");
      if ((fd=open(password_file,O_WRONLY|O_CREAT,0600))<0) {
	printf ("Error creating password file %s !\n",password_file);
	return -4;
      }
      close(fd);
      chown(password_file,0,0);
      chmod(password_file,0600);
      if ((st=fopen(password_file,"r+"))==NULL) {
	printf("Error opening created password file %s !\n",password_file);
	return -6;
      }
      found=0;
    } else {
      while (fgets(buffer,128,st)) { 
	found=0;
	for (i=0;i<=6;i++) {
	  if (buffer[i]==':' && argv[2][i]==0) {
	    found=1;
	    break;
	  }
	  if (buffer[i]!=argv[2][i])
	    break;
	}
	if (found)
	  break;
      }
    }
    if (!found) {      
      sprintf (buffer,"%s:%s:%s:::::\n",argv[2],crypt(argv[3],salt),argv[4]);
      fputs(buffer,st);
      fclose(st);
      printf("Added user %s, password %s, name %s\n",argv[2],argv[3],argv[4]);
      return 0;
    } else {
      // User already exist!
      // I have no time to add code to modify previous user data
      printf("User %s already exist!\n",argv[2]);
      fclose(st);
      return -6;
    }
  } else {
    // dpaccess_pass program called by no-root no-nobody user!
    printf("Sorry, you have no permissions to run this program!\n");
  }
  return -5;
}

